/*
 * Copyright (c) 1994
 *
 * Bettina Schnor,  Stefan Stille,  Henrik Carlsson
 *
 * TU Braunschweig, Germany
 * Institute for Operating Systems and Computer Networks
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that this copyright
 * notice appears in all copies.  The University of Braunschweig
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */


/*
 * $Id: loaddaemon.c,v 1.2 1994/02/28 18:57:33 carlsson Exp $
 * $Log: loaddaemon.c,v $
 * Revision 1.2  1994/02/28  18:57:33  carlsson
 * Searching for bugs that doesn't exist.
 *
 * Revision 1.2  1994/02/28  18:57:33  carlsson
 * Searching for bugs that doesn't exist.
 *
 * Revision 1.1  1993/12/21  10:17:15  carlsson
 * The following global variables were added:
 * 	idle_host	- Max load for a host to be considered idle.
 * 	loadpolling	- TRUE if loaddaemon shall answer loadpolling requests.
 * 	max_long	- Max allowed long jobs for host to be considered 
 *			  available.
 * 	long_jobs	- Number of long jobs executing at host (Pointer
 *			  to shared memory).
 *
 * The following new functions were added:
 * 	close_all()		- Closes all resources allocated by the
 *				  loaddaemon.
 * 				  Used as callback function when initializing
 *				  the child process
 * 				  handling routines. It is called before a
 *				  new child process is
 * 				  started.
 * 	readhostdatabase()	- Reads host and system configuration.
 *
 * The following functions were changed:
 * 	calculate()	- If the number of executing long jobs at host
 *			  is greater than max_long
 * 			  the host is considered as inavailable.
 * 			  The load from long jobs is subtracted from the
 *			  total load (rungue -= (*long_jobs)
 * 	main()		- The hardcoded names in calls to getservbyname 
 *			  were changed to DAEMOMN_SERVICE
 * 			  and LOMA_SERVICE.
 * 			  The use of the child process handling routines 
 *			  were introduced to start and restart
 * 			  the loadmanager and the rexecmanager.
 * 			  The messages CONFIG_READ_SYS and CONFIG_READ_HOST
 *			  were introduced in order to answer
 * 			  the requests to reread system and host
 *			  configuration respectively.
 * 			  The message KILL_REQUEST was added to be able
 *			  to exit correctly.
 *
 * Revision 1.0  1993/12/02  14:49:32  stille
 * Initial revision
 */

#include "system.h"
#include "children.h"
#include "yalb_config_db.h"
#include "reconfig.h"

#include <math.h>
#include <stdio.h>
#include <kvm.h>
#include <nlist.h>
#include <signal.h>
#include <vm/anon.h>
#include <sys/param.h>
#include <sys/dk.h>
#include <sys/vmmeter.h>
#include <sys/resource.h>
#include <net/route.h>
#include <netinet/in_pcb.h>
#include <sys/domain.h>
#include <sys/socketvar.h>
#include <sys/socket.h>
#include <sys/protosw.h>
#include <sys/proc.h>
#include <sys/user.h>
#include <sys/stat.h>
#include <sys/unpcb.h>
#define KERNEL
#include <sys/file.h>
#undef KERNEL
#define MAXSYMBOL 10
#define MAXWIDTH 10
#define TRUE 1
#define FALSE 0

#ifndef UPDATE
#define UPDATE 5                /* Update interval for load information (in sec) */
#endif /* !UPDATE */

#ifndef SECONDS
#define SECONDS 30                  /* Smoothing interval (not needed) */
#endif /* !SECONDS */

char myname[] = "loaddaemon";
struct nlist symbole[MAXSYMBOL];  /* List of symbols to be read from kernel */ 
kvm_t *mptr;                      /* Pointer to kmem */
float idle_host;                  /* Max load for a host to be considered idle */
int loadpolling;                  /* Should loaddaemon answer loadpolling requests */
int max_long;                     /* Max long lived jobs allowed on host for considering host available */
float arch_factor;                /* Architecture-dependent speed-factor */
float system_factor1,             /* Architecture-dependent system-factor for low system load */
  system_factor2;                 /* Architecture-dependent system-factor for high system load */
int send;                         /* Flag for sending load information to remote hosts */
rpc_message info;                 /* Message body for load broadcast */
int first = 1;                    /* Flag for first run */
static float old_load;            /* Old load index needed to
				   * determine whether broadcast of new
				   * load index is needed */
static int old_avail;             /* Old availability flag needed to
				   * determine whether broadcast of new
				   * load index is needed */
struct timeval *last_update,      /* Timestamp for last update */
  incoming;                       /* Timestamp for incoming requests */
int sock;                         /* Descriptor for UDP-socket */
struct sockaddr_in  remote;       /* Broadcast-Address of remote socket */
char host_name[15];               /* Name of local host  */
float *local_load;                /* Local load for shared memory */
int *availability;                /* Local availability for shared memory */
float *terminated;                /* Number of terminated jobs since last update of load-information */
int *temp_imported;               /* Number of jobs imported since last update of load-information */
int *long_jobs;                   /* Number of long jobs executing at host */

static int shared_mem;                   /* File descriptor for shared memory segment */
caddr_t addr_time;                /* Address of shared memory segment */

/*
 * close_all closes all open files, sockets, kvm, syslog  and 
 * shared memory segments.
 * It is used as the call back function in call to init_children
 * and will be called every time a children is started or restarted.
 */

static void close_all(void)
{
	(void)munmap(addr_time, (size_t) (sizeof(struct timeval) + sizeof(float) + 2*sizeof(int) + sizeof(float)));
	close(sock);
	close(shared_mem);
	kvm_close(mptr);
	closelog();
}

/*
 * Authenticates a given uid and returns TRUE if authenticated, FALSE
 * otherwise 
 */

int authenticate(what)
auth_request *what;
{
  struct proc *process;          /* Pointer to proc structure of given process */
  struct user *user;             /* Pointer to user structure of given process */
  struct file ofile;             /* File structure of a kernel object */
  struct file *fileptr;          /* Pointer to above mentioned structure */
  struct socket so;              /* Socket structure */
  struct protosw proto;          /* Socket protocol structure */
  struct domain dom;             /* Domain of given socket */
  struct inpcb pcb;
  struct ucred credits;          /* Process credentials */
  int count=0;

  
  /*
   * Fortunately, we have the kvm routines to read certain infomation
   * from the kernel. First, we read the process-structure of the
   * calling process, if any. 
   */
  

  while ((process = kvm_getproc(mptr, what->proc_nr)) == NULL) {
    count++;
    if (count == 10){
      return FALSE;
    }
  }
  
  /*
   * Try to read the credentials of the calling process
   */
  
  if(kvm_read(mptr, process->p_cred, (char *) &credits, sizeof (credits)) != sizeof (credits)){
    return FALSE;
  }
  
  if(credits.cr_uid != what->euid ||
     credits.cr_gid != what->egid ||
     credits.cr_ruid != what->ruid ||
     credits.cr_rgid != what->rgid){
    return FALSE;
  }
  
  /*
   * Now let's proof the user groups sent by the calling process
   */
  
{
  int count = 0;
  
  for (;credits.cr_groups[count] >= 0; count++) 
    if(credits.cr_groups[count] != what->groups[count]){
      return FALSE;
    }
}


/*
 * Next, we read its user structure from the kernel
 */

if((user = kvm_getu(mptr, process)) == NULL){
  return FALSE;
}


/*
 * Now, we should be able to read the ofile structure for the socket
 * on which the request occured. First, we consider the easy case
 * that the file structure for the socket is in the user-area
 * itself. This is indicated by a socket number less than
 * NOFILE_IN_U 
 */

if (what->sock_nr < NOFILE_IN_U) {
  if(kvm_read(mptr, user->u_ofile_arr[what->sock_nr], (char *) &ofile, sizeof (struct file)) != sizeof (struct file)){
    return FALSE;
  }
}
else
  
  /*
   * Consider the slightly complicated case that the file structure is
   * in dynamically allocated storage. First read the appropriate
   * address in kernel... 
   */
{
  if (kvm_read(mptr, (user->u_ofile + ((what->sock_nr - NOFILE_IN_U) * sizeof (struct file *))), 
	       (char *) &fileptr, sizeof (struct file *)) != sizeof (struct file *)){
    return FALSE;
  }
  
  /*
   * ... then get the file structure
   */
  
  if (kvm_read(mptr, fileptr, (char *) &ofile, sizeof (struct file)) != sizeof (struct file)){
    return FALSE;
  }
}

/*
 * Now we can consider the type (file/socket) of the kernel object.
 * If not a socket, somebody tried to cheat us !
 */

if (ofile.f_type != DTYPE_SOCKET){
  return FALSE;
}

/*
 * Read socket structure 
 */

if (kvm_read(mptr, ofile.f_data, &so, sizeof (so)) != sizeof (so) || so.so_pcb == NULL){
  return FALSE;
}

/*
 * Consider the protocol data structure and the domain of the socket
 */

if (kvm_read(mptr, so.so_proto, &proto, sizeof (proto)) != sizeof (proto)){
  return FALSE;
}

if (kvm_read(mptr, proto.pr_domain, (char *) &dom, sizeof (struct domain)) != sizeof (struct domain) ||
    dom.dom_family != PF_INET){
  return FALSE;
}

/*
 * Now we finally (!!) read the inpcb structure which contains the port
 * of the socket
 */

if (kvm_read(mptr, so.so_pcb, (char *) &pcb, sizeof (pcb)) != sizeof(pcb) ||
    pcb.inp_lport != what->port){
  return FALSE;
}


/*
 * Given port and user-id are correct!
 */

return TRUE;

}







void readhostdatabase(void)
{
	host_config local_host_config;
	system_config local_system_config;
	int linenr;

	switch( open_config(&linenr) ) {
	      case ERR_CFDB_OPEN:
		fprintf(stderr, "Cant open configuration file\n");
		exit(-1);
		break;
	      case OK_CFDB:
		break;
	      default:
		fprintf(stderr, "Error at line %d in configuration file", linenr);
		exit(-1);
		break;
	}
	if( read_host_config(&local_host_config, host_name) != OK_CFDB ||
	   local_host_config.choose != TRUE ) {
		fprintf(stderr, "This host is not configuered as part of Rexec-System.\nUse yalb_host_config to configure host.\n");
		exit(-1);
	}
	if( read_system_config(&local_system_config) != OK_CFDB ) {
		fprintf(stderr, "Cant read system configuration.\n");
		exit(-1);
	}
	if( !is_missval(TYPE_IDLE_HOST, (void *)&local_system_config.idle_host) ) {
		idle_host = local_system_config.idle_host;
	}
	else {
		idle_host = DEFAULT_IDLE_HOST;
	}
	if( !is_missval(TYPE_LOADPOLLING, (void *)&local_system_config.loadpolling ) ) {
		loadpolling = local_system_config.loadpolling;
	}
	else {
		loadpolling = DEFAULT_LOADPOLLING;
	}
	if( !is_missval(TYPE_ALPHA, (void *)&local_host_config.alpha) ) {
		arch_factor = local_host_config.alpha;
	}
	else {
		arch_factor = DEFAULT_ALPHA;
	}
	if( !is_missval(TYPE_SYS1, (void *)&local_host_config.sys1) ) {
		system_factor1 = local_host_config.sys1;
	}
	else {
		system_factor1 = DEFAULT_SYS1;
	}
	if( !is_missval(TYPE_SYS2, (void *)&local_host_config.sys2) ) {
		system_factor2 = local_host_config.sys2;
	}
	else {
		system_factor2 = DEFAULT_SYS2;
	}

	if( !is_missval(TYPE_MAX_LONG, (void *)&local_host_config.max_long) ) {
		max_long = local_host_config.max_long;
	}
	else {
		max_long = DEFAULT_MAX_LONG;
	}
	
	close_config();
}











/*
 * Procedure to broadcast load-information and save old load-information
 */


void broadcast_load()

{
#if 0  
  *last_update = incoming;
#else
  /* XXX: check rc */
  gettimeofday (last_update, (struct timezone *) 0);
#endif

  if (sendto(sock, &info , sizeof info, 0,(struct sockaddr *)&remote, sizeof remote) < 0)
    syslog(LOG_ERR, "Load-Daemon: Error broadcasting load_index: %m");
  
  /*
   * Save last load information broadcasted
   */
  
  old_avail = info.data_avail;
  old_load = info.data_load;
}

/*
 * Procedure to send back the current load to the sender only
 */

void send_load(from)
struct sockaddr_in *from;
{
  
  *last_update = incoming;
  
  if (sendto(sock, &info , sizeof info, 0,(struct sockaddr *)from, sizeof *from) < 0)
    syslog(LOG_ERR, "Load-Daemon: Error broadcasting load_index (%m)");
  
  /*
   * Save last load information broadcasted
   */
  old_avail = info.data_avail;
  old_load = info.data_load;
}

#ifndef USE_AVENRUN
/*
 * Function for handling SIGALARM
 *
 * Gathers statistics and calculates the load index
 * Sets new timer at end of execution
 * 
 */

void calculate()
{
  static int ret;                           /* Return value of function calls */
  static struct vmtotal itotal;             /* For determining the
					     * local runqueue (CPU) */
  static struct anoninfo swapinfo;          /* Swap info from kernel */
  static long cptime[CPUSTATES];            /* Ticks per CPU-state */
  static long oldcptime[CPUSTATES];         /* Previous ticks per CPU-state */
  static float cpuusage[CPUSTATES];         /* Averaged CPU usage */
  static float runqueue = 0.0;              /* Averaged runqueue (CPU) */
  int i;                                    /* Counter */
  static int broadcast;                     /* Flag whether broadcast is to be forced */
  static float smoothing;                   /* Used for smoothing the runqueue */
  struct timeval tv;


  if ((ret = kvm_read (mptr, symbole[0].n_value , &itotal, sizeof itotal)) != sizeof itotal){
    syslog(LOG_ERR, "Load-Daemon: Error reading runqueue from kernel (%m)");
  }
  
  if ((ret = kvm_read (mptr, symbole[1].n_value , &swapinfo, sizeof swapinfo)) != sizeof swapinfo){
    syslog(LOG_ERR, "Load-Daemon: reading swapinfo from kernel (%m)");
  }
  if ((ret = kvm_read (mptr, symbole[2].n_value , &cptime, sizeof cptime)) != sizeof cptime){
    syslog(LOG_ERR, "Load-Daemon: reading cptimeinfo from kernel (%m)");
  }
  
  if (first){
    for (i=0 ; i<CPUSTATES ; i++)
      oldcptime[i]=cptime[i];
    old_load = 10;    /* force broadcast for load-index */
    smoothing = (float) exp(-(double)UPDATE/(double)SECONDS);
  }
  
  info.data_avail = TRUE;
  
  /*
   * Examine swap space available 
   */
  
/*  if (((float)swapinfo.ani_free / (float)swapinfo.ani_max) < (SWAP_AVAIL / 100.0)){
    info.data_avail = FALSE;
  }
*/
  /*
   * Examine swapped processes
   */
  
  if (itotal.t_sw > 0){
    info.data_avail = FALSE;
  }
  
  if( *long_jobs >= max_long ) {
	  info.data_avail = FALSE;
  }

  /*
   * Update availability in shared memory segment
   */
  
  *availability = info.data_avail;
  
  /*
   * Calculate average CPU-usage for all CPU-states (This could be
   * restricted to system time only, since no other value is needed
   * for the load-index). That's the way I do it now!
   */
  
    cpuusage[SYSTEM] = (cptime[SYSTEM] -  oldcptime[SYSTEM]);
    oldcptime[SYSTEM] = cptime[SYSTEM];
  
  /*
   * Calculate average runqueue (CPU)
   */
  
  runqueue = (smoothing * runqueue + (float) itotal.t_rq * (1.0 - smoothing));
  
  /*
   * Check whether any job has been imported or terminated since the
   * last calculation of the load-index. 
   */
  
  broadcast = FALSE;

  if( (*long_jobs) > 0 ) {
	  runqueue -= (*long_jobs);
  }
  if (*terminated > 0){
    runqueue -= (*terminated);
    *terminated = 0;
    broadcast = TRUE;
  }
  if (*temp_imported > 0) {
    runqueue += 0.6 * (*temp_imported);
    *temp_imported = 0;
    broadcast = TRUE;
  }
  
  if(runqueue < 0.0)
    runqueue = 0.0;
  
  /*
   * Calculate load index
   */
  
  info.data_load = 1.0 + runqueue;
  
  if (cpuusage[SYSTEM] > 50.0)    
    info.data_load += system_factor1;
  
  if (cpuusage[SYSTEM] > 100.0)    
    info.data_load += system_factor2 * ((int)((cpuusage[SYSTEM] - 50)/50));
  
  info.data_load *= arch_factor;
  
  /*
   * Put local load into shared memory segment
   */
  
  *local_load = info.data_load;
  
  /*
   * If any job has finished or being started since last calculation
   * of the load-index, we force a load-broadcast 
   */
  
  /*
   * Otherwise, determine whether broadcast is needed on request of
   * any load-manager
   */
  
  if (((old_load / info.data_load < (1.0 - (float)LOAD_CHANGE/100.0)) ||
       (old_load / info.data_load > (1.0 + (float)LOAD_CHANGE/100.0))) ||
      (old_avail != FALSE && info.data_avail == FALSE))
    broadcast = TRUE;
  
  if (((old_load < idle_host && info.data_load > idle_host) || 
       (old_load > idle_host && info.data_load < idle_host)) ||
      (old_avail == FALSE && info.data_avail != FALSE))
    broadcast = TRUE;
  
  /*
   * send at least one loadinfo packet within the invalidateion interval:
   */

  if (gettimeofday (&tv, (struct timezone *) 0) >= 0 &&
      tv.tv_sec - last_update->tv_sec >= (MAX_MISSING_UPDATES - 2) * UPDATE)
      broadcast = TRUE;
  
  if (broadcast)
    broadcast_load();

  /*
   * Set new timer if not first call
   */
  
  if (first){
    first = FALSE;
    return;
  }
  alarm(UPDATE);
  
}
#else
/*
 * Function for handling SIGALARM
 *
 * Gathers statistics and calculates the load index
 * Sets new timer at end of execution
 * 
 */

void calculate()
{
  static int ret;                           /* Return value of function calls */
  static long aven[3];
  int i;                                    /* Counter */
  static int broadcast;                     /* Flag whether broadcast is to be forced */
  
  if ((ret = kvm_read (mptr, symbole[0].n_value , aven, sizeof(aven))) != sizeof(aven)){
    syslog(LOG_ERR, "Load-Daemon: reading runqueue from kernel (%m)");
  }
  
  /*
   * Update availability in shared memory segment
   */
  
  *availability = TRUE;

  /*
   * Calculate basic load index...
   */
  
  info.data_load = (float) aven[0] / 256.0;
  info.data_avail = TRUE;

  /*
   * Check whether any job has been imported or terminated since the
   * last calculation of the load-index. 
   */
  
  broadcast = FALSE;

  if( (*long_jobs) > 0 ) {
	  info.data_load -= (*long_jobs);
  }
  if (*terminated > 0){
    info.data_load -= (*terminated);
    *terminated = 0.0;
    broadcast = TRUE;
  }
  if (*temp_imported > 0) {
    info.data_load += 0.6 * (*temp_imported);
    *temp_imported = 0;
    broadcast = TRUE;
  }
  
  if(info.data_load < 0.0)
    info.data_load = 0.0;
    
  info.data_load += 1;
  info.data_load *= arch_factor;

  /*
   * Put local load into shared memory segment
   */
  
  *local_load = info.data_load;
  
  /*
   * If any job has finished or being started since last calculation
   * of the load-index, we force a load-broadcast 
   */
  
  if (broadcast)
    broadcast_load();

  else {
    
    /*
     * Otherwise, determine whether broadcast is needed on request of
     * any load-manager
     */
    
    send = FALSE;
    
    if (((old_load / info.data_load < (1.0 - (float)LOAD_CHANGE/100.0)) ||
	 (old_load / info.data_load > (1.0 + (float)LOAD_CHANGE/100.0))) ||
	(old_avail != FALSE && info.data_avail == FALSE))
      send = TRUE;
    
    if (((old_load < idle_host && info.data_load > idle_host) || 
	 (old_load > idle_host && info.data_load < idel_host)) ||
	(old_avail == FALSE && info.data_avail != FALSE))
      send = TRUE;
  }

  /*
   * Set new timer if not first call
   */
  
  if (first){
    first = FALSE;
    return;
  }
  alarm(UPDATE);
  
}
#endif /* !USE_AVENRUN */

int main(argc, argv)
int argc;
char *argv[];
{
  int ret = 0;                      /* Return value of function calls */
  struct hostent *local_address;    /* Internet-Address of local host  */
  int ldaemon;                      /* Port of load-daemon */
  int lmanager;                     /* Port of load-manager */
  struct servent *get_ports;        /* Structure to get port of load-daemon and load-manager */
  struct sockaddr_in local;         /* Address of local socket */
  struct sockaddr_in from;          /* Who sent the authentication request? */
  int fromlen = 
    sizeof (struct sockaddr_in);      /* Length of address */
  rpc_message request;              /* Buffer for incoming messages */ 
  auth_request a_request;           /* Authentication-request of remote rexec manager */
  auth_reply a_reply;               /* Reply to above request */
  int length=15;                    /* Length of local hostname  */
  
  int lmanager_id;                  /* Id of loadmanager returned from start_child() */
  int rmanager_id;                  /* Id of rexecmanager returned from start_child() */

  char path[MAX_PATH_LEN];



  /*
   * Install the function stop_children() with on_exit() so that it will be
   * called when the loaddaemon exits
   */

  on_exit(stop_children, 0);

  
  /*
   * Initialize system
   */

  OPENLOG(myname);

  syslog (LOG_WARNING, "Load-Daemon: starting up.");

  /*
   * Get host name and Internet-Address
   */
  
  if (gethostname(host_name,length) != 0)
    exit(1);
  
  if (!(local_address = gethostbyname(host_name)))
    exit(1);
  
  
  /*
   * Read host specific data from config-file
   */
  
  readhostdatabase();
  
  /*
   * Read kernel memory offsets for the kernel symbols to be read
   */ 
#ifndef USE_AVENRUN  
  symbole[0].n_name = "_total"; 
  symbole[1].n_name = "_anoninfo";
  symbole[2].n_name = "_cp_time";
  symbole[3].n_name = (char *)0;
#else  
  symbole[0].n_name = "_avenrun";
  symbole[1].n_name = (char *)0;
#endif /* !USE_AVENRUN */
  
  /*
   * Open kmem for reading 
   */
  
  mptr = kvm_open((char *)0 , (char *)0  , (char *)0 ,O_RDONLY, (char *)0 );
  
  if (!mptr) {
    syslog(LOG_ERR, "Load-Daemon: Could not open /dev/kmem  : %m");
  exit(1);
  }
  
  /*
   * Read memory offsets 
   */
  
  if(kvm_nlist (mptr, symbole) == -1){
    syslog(LOG_ERR, "Load-Daemon: Illegal Symbol in nlist-structure: %m");
    exit(1);
  }

  /*
   * Create shared memory for timestamps
   */
    
  if (mkdir(YALB_PATH,0755) < 0 && errno != EEXIST) {
    syslog(LOG_ERR, "Load-Daemon: YALB-directory not creatable! : %m");
    exit(-1);
  }
  
  if (chdir(YALB_PATH) < 0) {
    syslog(LOG_ERR, "Load-Daemon: YALB-directory not accessable! : %m");
    exit(-1);
  }

#ifdef ROOT_OWNER
  shared_mem = open(SHARED,O_RDWR | O_CREAT, 0600);
#else
  shared_mem = open(SHARED,O_RDWR | O_CREAT, 0666);
#endif


  if (shared_mem < 0){
    syslog(LOG_ERR, "Load-Daemon: Shared memory not available : %m");
    exit(1);
  }  

  
  ftruncate (shared_mem, 4096);

  addr_time = mmap ((caddr_t)0, (size_t) (sizeof(struct timeval) + sizeof(float) + 2*sizeof(int) + sizeof (float) + sizeof(int)), 
		    PROT_WRITE | PROT_READ, MAP_SHARED, shared_mem, (off_t)0 );   
  
  if (addr_time == (caddr_t)-1)
    syslog(LOG_ERR, "Load-Daemon: getting shared segment (%m)");
  
  last_update = (struct timeval *) addr_time;
  local_load = (float *) (addr_time + sizeof(struct timeval));
  availability = (int *) (addr_time + sizeof(struct timeval) + sizeof(float));
  temp_imported = (int *) (addr_time + sizeof(struct timeval) + sizeof(float) + sizeof(int));
  terminated = (float *) (addr_time + sizeof(struct timeval) + sizeof(float) + 2*sizeof(int));
  long_jobs = (int *) (addr_time + sizeof(struct timeval) + sizeof(float) + 2*sizeof(int) + sizeof(float));

  /*
   * Get port number for services "ldaemon" and "lmanager"
   */
  
  get_ports = getservbyname(DAEMON_SERVICE, "udp");
  
  ldaemon = get_ports->s_port;
  
  get_ports = getservbyname(LOMA_SERVICE, "udp");
  
  lmanager = get_ports->s_port;
  
  /*
   * Open Internet-Domain-Socket for sending/receiving 
   */
  
  if (argc == 2 && ! strcmp (argv [1], "-i"))
    /* pray for a miracle... */
    sock = 0;
  else
    sock = socket(AF_INET, SOCK_DGRAM, 0);
  
  if (sock < 0) 
  {
    syslog(LOG_ERR, "Load-Daemon: opening datagram socket (%m)");
    exit(1);
  }
  
  /*
   * Bind name to local socket
   */
  
  local.sin_family = AF_INET;
  local.sin_port = ldaemon;
  local.sin_addr.s_addr = INADDR_ANY;  
  
  if( bind(sock, (struct sockaddr *)&local, sizeof local) < 0)
  {
    syslog(LOG_ERR, "Load-Daemon: binding local socket (%m)");
    exit(1);
  }  
  
  /*
   * Get broadcast address 
   */
  
  bcopy((char *)local_address->h_addr, (char *)&remote.sin_addr, local_address->h_length);
  
  /* Local hack !! */
  remote.sin_addr.s_addr = (remote.sin_addr.s_addr & BROADCAST_MASK) | ~BROADCAST_MASK;
  
  remote.sin_family = AF_INET;
  remote.sin_port = lmanager;
  
  /*
   * Init load index
   */
  
  calculate();
  
  /*
   * Init message header for load broadcast
   */
  
  info.kind = LOAD_BROADCAST;
  info.magic_number = MAGIC;
  
  /*
   * Init message header for authentication message
   */
  
  a_reply.kind = AUTH_REPLY;
  a_reply.magic_number = MAGIC;
  
  /*
   * Install signal driver for SIGALARM
   */
  
  signal(SIGALRM,calculate);
  
  /*
   * Schedule first SIGALRM
   */
  
  alarm(UPDATE);
  
  /*
   * Force first load broadcast 
   */
  
  gettimeofday(&incoming, (struct timezone *)0);
  
  broadcast_load();
  

  /*
   * Initialize the child process handling functions
   */

  init_children(close_all);

  /* 
   * Start the daemons
   */

  sprintf(path, "%s/%s", YALB_DAEMON, LMANAGER);
  lmanager_id = start_child(path);
  sprintf(path, "%s/%s", YALB_DAEMON, RMANAGER);
  rmanager_id = start_child(path);




  /*
   * ... and now: Listen for incoming requests on socket 
   * and reply if necessary (main loop)
   */
  
  for(;;){
    
    /*
     * Listen to socket
     */
    
    if(recvfrom(sock, &request, sizeof request,0, &from, &fromlen) < 0){

      if(errno != EINTR) {
	syslog(LOG_ERR, "Load-Daemon: listening to socket (%m)");
      }
      else
	continue;
    }
    
    /*
     * Check incoming packet for magic number (poor authentication !)
     */
    
    if ((request.kind == LOAD_POLL) && (request.magic_number == MAGIC)){
      
      /*
       * Magic number and message type ok, so get new timestamp
       */
      
      if (loadpolling){
	gettimeofday(&incoming, (struct timezone *)0);
      
	/*
	 * If last request is older than UPDATE seconds and send is true
	 * broadcast new load-index and set new timestamp
	 */
      
	if (((incoming.tv_sec - last_update->tv_sec) >= UPDATE) && send)
	  broadcast_load();
      }
      continue;
    }
    
    if ((request.kind == LOAD_REQUEST) && (request.magic_number == MAGIC)){
      
      /*
       * Magic number and message type ok, so get new timestamp
       */
      
      gettimeofday(&incoming, (struct timezone *)0);
      
      /*
       * If last request is older than UPDATE seconds and send is true
       * broadcast new load-index and set new timestamp
       */
      
      send_load(&from);
      
      continue;
    }
    
    if ((request.kind == BROADCAST_FORCE) && (request.magic_number == MAGIC)){
      
      /*
       * Magic number and message type ok, so get new timestamp
       */
      
      gettimeofday(&incoming, (struct timezone *)0);
      
      /*
       * Broadcast new load-index and set new timestamp
       */
      
      broadcast_load();
      
      continue;
    }
    
    /*
     * Hey! We received an authentication request from a remote rexec
     * manager! We do our best to prevent him from fraud...
     */

    if( (request.kind == CONFIG_READ_SYS) && (request.magic_number == MAGIC) ) {
	    
	    readhostdatabase();

	    /*
	     * Send signal to children
	     */
	    kill_child(lmanager_id, SIG_CONFIG_READ);
	    kill_child(rmanager_id, SIG_CONFIG_READ);
	
	    continue;
    }

    if( (request.kind == CONFIG_READ_HOST) && (request.magic_number == MAGIC) ) {
    
	    readhostdatabase();

	    /*
	     * Reply to the read request so that the caller doesn't try to start us
	     */
	    reply_to_config_read(sock, (struct sockaddr *)&from);

	    continue;
    }

    if( (request.kind == KILL_REQUEST) && (request.magic_number == MAGIC) ) {
	    exit(0);
    }
    
    else {
      
      int mask;
      
      if ((request.kind == AUTH_REQUEST) && (request.magic_number == MAGIC)){
	
	a_request = *(auth_request *) &request;

	mask=sigblock(sigmask(SIGALRM));
	a_reply.ok = authenticate(&a_request);
	sigsetmask(mask);

	if(sendto(sock, (char *) &a_reply, sizeof (auth_reply), 0, &from, fromlen) < 0 ) {
	  syslog(LOG_ERR, "Load-Daemon: sending authentication reply (%m)");
	}
      }
      
      /*
       * If not 'authenticated', discard message and loop again
       */
      
    }
    
  }
  
  /* Never reached */
  
  exit(0);
  
}

