/*
 * Copyright (c) 1994
 *
 * Bettina Schnor,  Stefan Stille,  Henrik Carlsson
 *
 * TU Braunschweig, Germany
 * Institute for Operating Systems and Computer Networks
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that this copyright
 * notice appears in all copies.  The University of Braunschweig
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

/*
 * $Id: remote.h,v 1.2 1993/12/15 08:21:47 carlsson Exp $
 * $Log: remote.h,v $
 * Revision 1.2  1993/12/15  08:21:47  carlsson
 * Moved the definitions of NORMAL_JOB and LONG_JOB to rexecbib.h.
 *
 * Revision 1.1  1993/12/14  19:03:36  carlsson
 * The field `jobkind' added to structure exec_data.
 * Two defines (NORMAL_JOB and LONG_JOB) which describes the possible 
 * values of `jobkind' added.
 *
 * Revision 1.0  1993/12/02  15:13:47  stille
 * Initial revision
 */

#ifndef _LOBA_SYSTEM
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/param.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ipc.h>
#include <sys/mman.h>
#include <netinet/in.h>
#include <sys/un.h>
#include <netdb.h>
#include <stdlib.h>
#include <sys/syscall.h>
#include <sys/wait.h>
#define MAGIC 180568                       /* Magic number used for (poor !) authentication */
#endif /* _LOBA_SYSTEM */

/*
 * Message types used in remote exec system
 */

#define REQUEST_FOR_EXEC 0                 /* Request for remote execution */
#define AUTHENTICATE     1                 /* Rexec-Manager demands authentication */
#define EXECUTE          2                 /* Message that contains data for execution */
#define EXEC_ERROR       3                 /* Error in executing command */
#define EXIT_STATUS      4                 /* Status of executed command returned by wait3() */
#define EXECUTION_OK     5                 /* Notification for remote master that remote execution is ok */
#define NO_EXECUTION     6                 /* Notification for remote master that remote execution is refused */

/*
 * Administrative stuff
 */

#define MAX_NUM_GROUPS 10                  /* Maximum number of groups a user can be in */
#define MAX_LEN 4096                       /* Maximum length of variable execution data */
#define ALIGN(ptr, type)    (type) (((long)(ptr)+3)&~3) /* Aligns pointer to next longword */

/*
 * Error codes returned from rexec-manager
 */

#define NO_ROOT_EXEC       1               /* Root execution not allowed */
#define NO_AUTHENTICATION  2               /* No authentication from local server */
#define INTERNAL_ERROR     3               /* Error internal to rexec-manager */
#define NO_FORK            4               /* Could not fork to execute command */

/*
 * Data Structures for remote execution
 */

typedef struct {
  int kind;                                /* Kind of message */
  int magic_number;                        /* Magic number for authentication */
  int length;                              /* Length of this message */
  int job;                                 /* Number of job on local site */
  u_short port;                            /* Port for stdin/stdout/stderr of remote process */
  int sock_nr;                             /* Socket on which the call was made */
  int proc_nr;                             /* Process number of requesting process */
  u_short ruid;                            /* Real user id */
  u_short euid;                            /* Effective user id */
  u_short rgid;                            /* Real group id */
  u_short egid;                            /* Effective group id */
  long umask;                              /* File creation mask */
  int ngroups;                             /* Number of groups */
  int groups[MAX_NUM_GROUPS];              /* Groups the user is in */
  int jobkind;                             /* Kind of job */
  char buf[MAX_LEN];                       /* Buffer for execution stuff (variable length)
					    * contains:
					    * - current work directory
					    * - file to execute (not necessarily absolute)
					    * - number of arguments (long value)
					    * - argument strings
					    * - number of environment strings (long value)
					    * - environment strings
					    */
} exec_data;

typedef struct {
  int kind;                                 /* Kind of message */
  int magic_number;                         /* Magic number for authentication */
  union {
    int errorcode;                          /* Errormessage to be sent to application */
    union wait state;                       /* Exit status from wait3() */
  } data;
#define r_error data.errorcode
#define r_status data.state
} rexec_reply;

typedef struct {
  int kind;                                 /* Kind of message */
  int magic_number;                         /* Magic number for authentication */
} notify;
