/*
 * Copyright (c) 1994
 *
 * Bettina Schnor,  Stefan Stille,  Henrik Carlsson
 *
 * TU Braunschweig, Germany
 * Institute for Operating Systems and Computer Networks
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that this copyright
 * notice appears in all copies.  The University of Braunschweig
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

/*
 * $Id: rexec.c,v 1.2 1994/02/24 15:19:52 carlsson Exp $
 * $Log: rexec.c,v $
 * Revision 1.2  1994/02/24  15:19:52  carlsson
 * Searching for bugs.
 *
 * Revision 1.1  1993/12/15  08:19:46  carlsson
 * Added the flags -f for getting the fastest host, and -l for executing
 * long jobs.
 *
 * Revision 1.0  1993/12/02  14:57:37  stille
 * Initial revision
 */

#include <stdio.h>
#include "lobalib.h"  
#include "rexecbib.h"

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif


static struct timeval vorher, nachher;
static float tdiff;
static FILE *stats;
/* char buff[100]; */
/* char command[100]; */

static be_quiet = 0;
static be_verbose = 0;


void exit_routine()
     
{
	exit_state state;
	
	while  (get_exit_status (&state)) {
		
		if (state.kind == INTERNAL){
			fprintf(stderr, 
	"Internal error: %d\nPlease report this to stille@ibr.cs.tu-bs.de", 
				state.data.errorcode);
			rexec_close(); 
			exit(-1);
		}
		else {
			rexec_close(); 
			exit(state.data.state.__w_retcode);
		}
		
	}
}

int main (argc, argv)
     int argc;
     char *argv[];
     
{
	
	request want;       /* Contains the wanted hosts */
	reply   get;        /* Contains the reply from load-manager */
	int ret; 
	struct hostent *name;
	char *addr;
	int i;
	struct timeval *last_update;
	float *local_load;
	int *availability;
	char hostname[20];
	int length = 20;
	caddr_t addr_time;
	char path[30];
	int fastest_host;
	int specified_host;
	int jobkind;
	int c;

	extern char *optarg;
	extern int optind;
	
	fastest_host = FALSE;
	specified_host = FALSE;
	jobkind = NORMAL_JOB;

	while( (c = getopt(argc, argv, "vqnh:fl")) != -1 ) {
		switch( c ) {
		case 'q':
		    be_verbose = ! (be_quiet = 1);
		    break;
		case 'v':
		    be_quiet = ! (be_verbose = 1);
		    break;
		case 'h':
		    strcpy(hostname, optarg);
		    specified_host = TRUE;
		    break;
		case 'f':
		    fastest_host = TRUE;
		    break;
		case 'l':
		    jobkind = LONG_JOB;
		    break;
		case 'n':
		    /* XXX (schoenfr) fix me: no input: */
		    (void) freopen ("/dev/null", "r", stdin);
		    break;
		case '?':
		    fprintf(stderr, 
    "usage: rexec [-v | -q] [-n] [-h <hostname> | -f] [-l] <command>\n");
		    exit(-1);
		}
	}

	if( (fastest_host && specified_host) || (argc - optind < 1) ) {
		    fprintf(stderr, 
	  "usage: rexec [-v | -q] [-n] [-h <hostname> | -f] [-l] <command>\n");
		exit(-1);
	}

	i = optind;

	if( specified_host ) {
		name = gethostbyname(hostname);
		
		if (!name) {
		    fprintf(stderr, "You specified an unknown host!\n");
			exit(-1);
		}
		
		bcopy((char *) name->h_addr , &get.hostlist[0].sin_addr, 
		      sizeof get.hostlist[0].sin_addr);
		
	}
	else {

		gethostname(hostname,length);
		
		if( fastest_host ) {
			want.kind = GET_FASTEST;
		}
		else {
			want.kind = GET_TOP; 
		}
		want.number = 1; 
		want.hostlist[0] = (struct sockaddr_in *)0;
		
		ret = request_init();
		
		if(ret != 0) {
		    fprintf(stderr, "System seems to be down !\n");
		    exit(1);
		}
    
		ret = get_hosts(&want,&get); 
    
		if(ret != 0) {
		    fprintf(stderr, "Error getting host: %d\n", ret);
		    exit(1);
		}
		
		name = gethostbyaddr((char *) &get.hostlist[0].sin_addr,
				   sizeof(get.hostlist[0].sin_addr), AF_INET);
		
		reset_system();
	}

	if (be_verbose)
	    fprintf(stderr, 
		    "Please report any errors to stille@ibr.cs.tu-bs.de !\n" );
	if (! be_quiet)
	    fprintf(stderr, "[Your job is exported to: %s]\n",name->h_name);
  
	if (rexec_init(INTERACTIVE) <0) {
	    fprintf(stderr, "Initialisation of rexec functions failed! \n\n");
		exit(1);
	}
	
	ret = rexecv (argv[i], &argv[i+1], &get.hostlist[0], jobkind);

	if (ret < 0) {
	    if (ret == COULD_NOT_CONNECT) {
			fprintf(stderr, "%s is unreachable.\n", name->h_name);
		}
		if (ret == CONNECTION_CLOSED) {
		    fprintf(stderr, "%s refuses execution.\n",
			    name->h_name);
		}
		if (ret == NO_RESOURCES)
		  fprintf(stderr, "Local Resources exhausted.\n");
		rexec_close(); 
		exit(-1);
	}

	exit_routine();
}  

