/*
 * Copyright (c) 1994
 *
 * Bettina Schnor,  Stefan Stille,  Henrik Carlsson
 *
 * TU Braunschweig, Germany
 * Institute for Operating Systems and Computer Networks
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that this copyright
 * notice appears in all copies.  The University of Braunschweig
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

/*
 * $Id: rexecbib.h,v 1.2 1993/12/15 10:05:24 carlsson Exp $
 * $Log: rexecbib.h,v $
 * Revision 1.2  1993/12/15  10:05:24  carlsson
 * Moved definitions of NORMAL_JOB and LONG_JOB to system.h.
 *
 * Revision 1.1  1993/12/15  08:20:58  carlsson
 * Moved the definitions of NORMAL_JOB and LONG_JOB to rexecbib.h.
 *
 * Revision 1.0  1993/12/02  15:04:26  stille
 * Initial revision
 */




/*
 * Flags given to rexec_init 
 */

#define INTERACTIVE 1                   /* Interactive execution. Only one Job at a time */
#define PARALLEL    2                   /* Execution of several parallel jobs at a time */


/*
 * Type of result returned by rexec_result()
 */

#define INTERNAL  1                     /* Error internal to the rexec-system */
#define EXECUTION 2                     /* Errorcode returned from remote execution */

#define MAX_ARGS 100                    /* Maximum number of arguments given to rexec() */

/*
 * Error-codes returned by rexec()
 */

#define NO_MORE_JOBS       -1        /* Number of pending jobs exceeded */
#define NO_RESOURCES       -2        /* System resources exhausted */
#define COULD_NOT_CONNECT  -3        /* Remote host unreachable */
#define CONNECTION_CLOSED  -4        /* Remote host closed connection */
#define NO_INIT            -5        /* System was not (successfully) 
					initialized using rexec_init() */
#define NO_JOBS_PENDING    -6	     /* No more jobs are pending on an
					rexec_wait() call */
#ifndef FATAL
#define FATAL		   -7            /* Fatal error */
#endif

/*
 * Error-code returned by rexec_init()
 */

#define ALREADY_INITED     1             /* Second call to rexec_init while system is already in use */

typedef struct {
  int kind;                              /* Kind of exit-status (INTERNAL/EXECUTION) */
  int job;                               /* Which Job exited? */
  union {
    int errorcode;                       /* Internal error */
    union wait state;                    /* Exit status from wait3() */
  } data;
} exit_state;


int rexec_init (/* int flag */);

int rexec(/* params */);

int rexecv(/* params */);


/*
 * Provides a clean exit from rexec-system. if any jobs are pending,
 * they are killed using SIGKILL.
 */

int rexec_close();



/*
 * Wait for a remote process to exit (PARALLEL mode only). This call
 * blocks until a process exited. On success (any job exited) the
 * function returns TRUE on failure NO_JOBS_PENDING (see above). If
 * the whole system is not initialized, it returns NO_INIT (see above)
 */

int rexec_wait();



/*
 * Get another exit-status from any remote process. Returns TRUE if
 * any exit-status is available, FALSE if not
 */

int get_exit_status (/* exit_state *exit_status; */);
