/*
 * Copyright (c) 1994
 *
 * Bettina Schnor,  Stefan Stille,  Henrik Carlsson
 *
 * TU Braunschweig, Germany
 * Institute for Operating Systems and Computer Networks
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that this copyright
 * notice appears in all copies.  The University of Braunschweig
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */


/*
 * $Id: rexecmanager.c,v 1.2 1994/02/28 19:21:27 carlsson Exp $
 * $Log: rexecmanager.c,v $
 * Revision 1.2  1994/02/28  19:21:27  carlsson
 * Looking for bugs that appears to be caused bu gcc!
 *
 * Revision 1.1  1993/12/21  10:34:27  carlsson
 * The global variable long_jobs was added. It contains the number 
 * of long jobs executing at host and is a pointer to shared memory.
 * The field long_job was added to the table procs. If it is TRUE, 
 * the corresponding process is a long job.
 *
 * The following functions were added:
 * 	readsysconfig()		- Read the system configuration.
 * 	handle_SIG_CONFIG_READ()- Handler for the signal SIG_CONFIG_READ
 *							(SIGUSR1).
 *
 * The following functions were changed:
 * 	handle_SIGCHLD()	- Statistics is written when a job ends.
 * 				  If the job was a long job the counter 
 *				  long_jobs is decreased.
 * 	main()			- The hardcoded names in call to 
 *				  getservbyname was changed to REMA_SERVICE
 *				  and DAEMON_SERVICE respectively.
 * 				  The new field jobkind in the 
 *				  REQUEST_FOR_EXEC message is checked. If 
 *				  it is true the job is marked as a long job
 * 				  and the counter long_jobs in shared memory
 *				  is increased.
 * 				  Statistics (timeouts and started jobs) 
 *				  were added.
 *
 * Revision 1.0  1993/12/02  14:55:37  stille
 * Initial revision
 */

#include <stdio.h>

#include "system.h"
#include "remote.h"
#include <sys/mman.h>

#include "yalb_stats_db.h"
#include "yalb_config_db.h"

#define TRUE 1
#define FALSE 0
#define SHORT_LIVED 3                      /* Age in seconds up to which a process is considered short-lived */

/* Default values for certain symbols ... */

#ifndef MAX_TRIES
#define MAX_TRIES 3              /* Number of tries to get an authentication */
#endif /* !MAX_TRIES */

#ifndef AUTH_TIMEOUT
#define AUTH_TIMEOUT 1000000     /* Timeout for authentication from loaddaemon (in usec) */
#endif /* !AUTH_TIMEOUT */

#ifndef DATA_TIMEOUT
#define DATA_TIMEOUT 2000000     /* Timeout for getting execution-data from remote site (in usec) */
#endif /* !DATA_TIMEOUT */

#ifndef MAX_SOCKS
#define MAX_SOCKS 64             /* Maximum number of sockets for rexec-manager */
#define MAXWIDTH 64
#else 
#define MAXWIDTH MAX_SOCKS
#endif /* !MAX_SOCKS */

#ifndef MAX_CONNECTS
#define MAX_CONNECTS 15                    /* Maximum number of connections for rexec-manager */
#endif /* !MAX_CONNECTS */

#define REFUSE         shutdown (socks[last].socket,2); \
  close (socks[last].socket);\
  socks[last].socket = -1;\
  imported--;

/*
 * Global variables
 */

extern char **environ;         /* Current environment */
int timeout;                   /* Indicates that a timeout occurs */
int last = 0;                  /* First free index in socks data-structure */
int imported = 0;              /* Number of imported jobs */
fd_set exceptfds;              /* Sockets on whichh to look for oob-data */
rexec_reply result;            /* Result to be sent to client */
notify notification;           /* Notification for remote master */
float *terminated;             /* Number of terminated jobs since last update of load-information */
int *temp_imported;            /* Number of jobs imported since last update of load-information */
int *long_jobs;                 /* Number of executing long jobs in shared mem */
char myname[] = "Rexec-manager";
#ifdef PROTOCOL
struct sockaddr_in from;         /* Who requested connection? */
char *file;                      /* Name of the file to be executed */
#endif

/*
 * Data needed for identifying processes is kept in a per-socket
 * data-structure
 */

struct {
	int proc_id;                   /* Number of associated process */
	struct timeval started;        /* Time the imported job was started */
	struct timeval ended;          /* Time the imported job was terminated */
	int long_job;                  /* TRUE if the process is a long job */
} procs[MAX_SOCKS];

/*
 * Contains active sockets
 */

struct {
	int socket;                 /* Active socket */
	struct sockaddr_in host;    /* For which host are we executing */
} socks[MAX_CONNECTS];

/*
 * Handle died or stopped children
 */

void handle_SIGCHLD ()
{
	union wait status;               /* Status of the child process */
	int pid;                         /* Process-id of child process */
	int sock;                        /* Socket belonging to process */
	int count;             
	struct rusage resourcen;         /* Resources needed by our child */
	float duration;                  /* Duration of command execution */
	int ret;
	
	/*
	 * Get pid of process, whose status has changed. We block on this
	 * because we do not want to miss an exiting child process
	 */
	
	while(((pid = wait3((int *)&status, 0, (struct rusage *) &resourcen)) < 0) && (errno == EINTR));
	
#ifdef PROTOCOL    
	{
		FILE *stats;
		
		if(stats =  fopen("/tmp/rexec.log","a")) {
			fprintf(stats,"Received SIGCHLD !\n");
			fclose(stats);
		}
	}
#endif
	

  /*
   * We are only interested in terminated processes
   */
	
	if((!WIFEXITED(status) && !WIFSIGNALED(status)) || pid == 0) {
#ifdef PROTOCOL    
		{
			FILE *stats;
			
			if(stats =  fopen("/tmp/rexec.log","a")) {
				fprintf(stats,"Job not exited. Process Id: %d\n", pid);
				fclose(stats);
			}
		}
#endif
		return;
	}
	
	for (sock = 0 ; (sock < MAX_SOCKS) && (procs[sock].proc_id != pid); sock++);
	
	/*
	 * If no entry in procs--structure, quit handling of SIGCHILD
	 */
	
	if(sock == MAX_SOCKS) {
#ifdef PROTOCOL    
		{
			FILE *stats;
			
			if(stats =  fopen("/tmp/rexec.log","a")) {
				fprintf(stats,"Job not in process-table, process-Id: %d\n", pid);
				fclose(stats);
			}
		}
#endif
		
		return;
	}
	/*
	 * Send exit-status of dead process back to caller, clear socks
	 * table entry and close associated socket
	 */
	
	gettimeofday(&procs[sock].ended, (struct timezone *)0);

	{
		/*
		 * Write statistics
		 */

		static stats_job_end job_end_info;

		job_end_info.proc_nr = procs[sock].proc_id;
		stats_write(STATS_TYPE_JOB_END, (stats_info)&job_end_info);
	}
	
	result.kind = EXIT_STATUS;
	result.r_status = status;
	
	while ((ret = send(sock, (char *)&result, sizeof (result), 0)) < 0  && (errno == EINTR));
	
	
#ifdef PROTOCOL    
	{
		FILE *stats;
		struct hostent *name;
		struct timeval time;
		char host[15];
		
		if(stats =  fopen("/tmp/rexec.log","a")) {
			name = gethostbyaddr((char *)&from.sin_addr.s_addr, sizeof (from.sin_addr.s_addr), AF_INET);
			gethostname(host, sizeof(host));
			gettimeofday(&time, (struct timezone *)0);
			fprintf(stats,"Sent to: %10s Command received: %s Bytes: %d\n", 
				name->h_name, file,  ret);
			fclose(stats);
		}
	}
#endif
	
	
	procs[sock].proc_id = -1;
	shutdown(sock,2);
	close(sock);
	FD_CLR(sock,&exceptfds);
	
	for(count = 0; (socks[count].socket != sock) && (count < MAX_CONNECTS); count++);
	
	/*
	 * If no entry in socks-structure, quit handling of SIGCHLD
	 */
	
	if(count == MAX_CONNECTS)
	  return;
	
	socks[count].socket = -1;
	
	imported--;
	
	/*
	 * Estimate load that was caused by terminated process and put it
	 * into shared memory...
	 */
	
	duration = (((double)procs[sock].ended.tv_sec - (double)procs[sock].started.tv_sec) + 0.000001 * ((double)procs[sock].ended.tv_usec - (double)procs[sock].started.tv_usec));
	
	/*
	 * The load of short lived processes is estimated by 0.6...
	 */
	
	if(duration > SHORT_LIVED) 
	  (*terminated)  += ((double) resourcen.ru_utime.tv_sec + 0.000001 * (double)resourcen.ru_utime.tv_usec) / duration;
	else
	  (*terminated) += 0.6;
	
	if( procs[sock].long_job ) {
		/*
		 * Decrease the number of long jobs in shared memory
		 */
		(*long_jobs)--;
	}
	
}

/*
 * Handle out-of-band data (propagate signal to appropriate process)
 */

void handle_SIGURG ()
{
	fd_set selectfds;               /* Temporary mask for select */
	int mask;                       /* Mask of blocked signals */
	int oob_socket;                 /* Socket from on which oob-data is received */
	char sig_char;                  /* We have to encode the signal in a single byte (oob-data!) */
	int ret;                        
	fd_set zerofds;
	
	/*
	 * Out-of-band-data is most important to us , so block all signals we
	 * can. 
	 */
	
	FD_ZERO(&zerofds);
	
	mask = sigblock (~0);
	
	selectfds = exceptfds;
	
	if(!bcmp((char *) &selectfds, (char *) &zerofds, sizeof (fd_set))) {
		sigsetmask(mask);
		return;
	}
	
	while ((ret = select(MAXWIDTH, (fd_set *)0, (fd_set *)0, &selectfds, (struct timeval *)0)) < 0 && errno == EINTR);
	
	if (ret < 0)  {
		syslog(LOG_ERR, "Error on select when receiving oob-data: %m");
		sigsetmask(mask);
		return;
	}
	
	/*
	 * We got the socket to listen to. Extract it from selectfds and read
	 * oob-data.
	 */
	
	for (oob_socket = 0; oob_socket < MAX_SOCKS; oob_socket++) {
		
		if(FD_ISSET(oob_socket, &selectfds)) {
			while (recv(oob_socket, &sig_char, sizeof (char), MSG_OOB) < 0) 
			  
			  switch (errno) {
				  
				case EWOULDBLOCK:
				  /*
				   * Urgent data not here yet. Since we do not use the sockets
				   * selected  upon for 'normal' data transfer, there is not way for
				   * us to block on a full buffer. So we simply wait for the data to
				   * be delivered to us. If any changes are made to this, we might
				   * have to change this and assure that the buffer is read so
				   * that the oob-data can be transmitted.
				   */
				  continue;
				  
				default:
				  
				  syslog(LOG_ERR, "Error receiving oob-data : %m");
				  sigsetmask(mask);
				  return;
				  
			  }
			
			/*
			 * Deliver signal to appropriate process
			 */
			
			ret = killpg(procs[oob_socket].proc_id, (long) sig_char);
			
		}
	}
	
	/*
	 * Reinstall old signal mask and exit
	 */
	
	sigsetmask(mask);
	return;
}

/*
 * Extract the transmitted array of strings and return an array of
 * char pointers. Used to extract the argv and envp for the new
 * process 
 */

static char **extract (strPtrPtr)
     char          **strPtrPtr;
{
	char *cp;                          /* Points to currently extracted string */
	char **vec;                        /* Array member (address of currently extracted string) */
	long  numStrings;                  /* Number of strings in buffer */
	char  **vecPtr;                    /* Pointer to resulting array of char pointers */
	
	/*
	 * First get the number of arguments
	 */
	
	cp = *strPtrPtr;
	numStrings = *(long *)cp;
	
	/*
	 * Allocate memory for the pointer array
	 */
	
	vecPtr = (char **)malloc((unsigned)((numStrings + 1) * sizeof(char *)));
	
	/*
	 * Now fill the pointer array with appropriate addresses
	 */
	
	
	cp += sizeof(long);
	for (vec = vecPtr; numStrings != 0; vec++, numStrings--) {
		*vec = cp;
		cp += strlen(cp) + 1;
	}
	
	/*
	 * Set terminating entry of pointer list
	 */
	
	*vec = (char *)0;
	
	/*
	 * Return new pointer for analyzing and pointer to the extracted
	 * char pointer array
	 */
	
	*strPtrPtr = cp;
	return(vecPtr);
}


void readsysconfig(void)
{
	system_config local_system_config;
	int linenr;
	
	switch( open_config(&linenr) ) {
	      case ERR_CFDB_OPEN:
		fprintf(stderr, "Cant open configuration file\n");
		exit(-1);
		break;
	      case OK_CFDB:
		break;
	      default:
		fprintf(stderr, "Error at line %d in configuration file", linenr);
		exit(-1);
		break;
	}
	
	if( read_system_config(&local_system_config) != OK_CFDB ) {
		fprintf(stderr, "Cant read system configuration.\n");
		exit(-1);
	}

	close_config();
	
	if( local_system_config.statistics == TRUE ) {
		stats_on();
	}
	else {
		stats_off();
	}
}




void handle_SIG_CONFIG_READ(void)
{
	readsysconfig();
}



/*
 * Function main
 */

int main ()
     
{
	struct servent *get_port;        /* Structure to get port of rexec-manager */
	int in_socket;                   /* Socket for accepting connections */
	int auth_server;                 /* Socket for authentication */
	struct sockaddr_in my_name;      /* Internet-name of local socket */
	struct sockaddr_in auth_name;    /* Internet-name of authentication service */
	auth_request request;            /* Request for authenticaton */
	auth_reply reply;                /* Reply to authentication request */
	fd_set main_read;                /* Mask for select on local socket */
	fd_set auth_fd;                  /* Mask for select on authentication socket */
	fd_set data_fd;                  /* Mask for select on data socket */
	int ret;                         /* Return-value of procedure-calls */
	int count;                       /* Counter */
	int shared_mem;                  /* File descriptor for shared memory segment */
#ifndef PROTOCOL
	struct sockaddr_in from;         /* Who requested connection? */
#endif
	struct sockaddr_in where;        /* Address to connect stdin etc. to */
	int fromlen = 
	  sizeof (struct sockaddr_in);   /* Length of above address */
	exec_data buf;                   /* Buffer for incoming execution-data */
	char *analyze;                   /* Used to parse variable part of exec-data */
	char *cwd;                       /* Pointer to current work directory */
	char newcwd[1024];               /* Buffer for automount-compatible path-name */
#ifndef PROTOCOL
	char *file;                      /* Name of the file to be executed */
#endif
	char **argv;                     /* List of arguments */
	char **envp;                     /* List of environment variables */
	int on = TRUE;                   /* Flag used to set in_socket into SO_REUSEADDR-mode */
	int authenticated;               /* Flag used to indicate whether authentication-message received */
	struct timeval auth_timeout;     /* Timeout-interval for authentication */
	struct timeval data_timeout;     /* Timeout-interval for receiving rexec_data */
	caddr_t addr_time;               /* Address of shared memory segment */
	int main_mask;                   /* Signalmask to be stored */
	
#define ERROR(no)   result.kind = EXEC_ERROR;\
	result.r_error = no;\ 
	  send(socks[last].socket, &result, sizeof (result), 0); 
	/*
	 * Open syslog
	 */
	
	OPENLOG(myname);
	
	/*
	 * Set timeout-interval for authentication
	 */
	
	auth_timeout.tv_sec = (long) AUTH_TIMEOUT / 1000000;
	auth_timeout.tv_usec = (long) AUTH_TIMEOUT - (auth_timeout.tv_sec * 1000000);
	
	/*
	 * Set timeout-interval for getting job data from remote site
	 */
	
	data_timeout.tv_sec = (long) DATA_TIMEOUT / 1000000;
	data_timeout.tv_usec = (long) DATA_TIMEOUT - (data_timeout.tv_sec * 1000000);
	
	/*
	 * Open Internet-Domain-Socket for receiving connections from applications
	 */
	
	in_socket = socket(AF_INET, SOCK_STREAM, 0);
	
	if (in_socket < 0) {
		syslog(LOG_ERR, "Rexec-Manager: Error opening stream socket : %m");
		exit(1);
	}
	
	/*
	 * Create shared memory for timestamps and local load
	 */
	
#ifdef ROOT_OWNER
	if ((mkdir(YALB_PATH,0500) < 0) && (errno != EEXIST)) {
		syslog(LOG_ERR, "Rexec-Manager: YALB-directory not creatable ! (%m)");
		exit(-1);
	}
#else
	if ((mkdir(YALB_PATH,0555) < 0) && (errno != EEXIST)) {
		syslog(LOG_ERR, "Rexec-Manager: YALB-directory not creatable ! (%m)");
		exit(-1);
	}
#endif
	
	if (chdir(YALB_PATH) < 0) {
		syslog(LOG_ERR, "Rexec-Manager: YALB-directory not accessable ! (%m)");
		exit(-1);
	}
	
#ifdef ROOT_OWNER
	shared_mem = open(SHARED,O_RDWR | O_CREAT, 0600);
#else
	shared_mem = open(SHARED,O_RDWR | O_CREAT, 0666);
#endif
	
	
	if (shared_mem < 0){
		syslog(LOG_ERR, "Rexec-Manager: Shared memory not available : %m");
		exit(1);
	}  
	
	ftruncate (shared_mem, 4096);
	
	addr_time = mmap ((caddr_t)0, (size_t) (sizeof(struct timeval) + sizeof(float) + 2*sizeof(int) + sizeof (float) + sizeof(int)), 
			  PROT_WRITE | PROT_READ, MAP_SHARED, shared_mem, (off_t)0 );   
	
	if (addr_time == (caddr_t)-1) {
		syslog(LOG_ERR, "Rexec-Manager: Error getting shared segment: %m");
		exit(1);
	}
	
	/*
	 * Place number of imported and terminated jobs into shared memory
	 */
	
	temp_imported = (int *) (addr_time + sizeof(struct timeval) + sizeof(float) + sizeof(int));
	terminated = (float *) (addr_time + sizeof(struct timeval) + sizeof(float) + 2*sizeof(int));
	long_jobs = (int *) (addr_time + sizeof(struct timeval) + sizeof(float) + 2*sizeof(int) + sizeof(float));
	*terminated = 0;
	*temp_imported = 0;
	*long_jobs = 0;
	

	readsysconfig();

	/*
	 * Install signal handler for SIG_CONFIG_READ (SIG_CONFIG_READ is defined in system.h)
	 */
	
	signal(SIG_CONFIG_READ, handle_SIG_CONFIG_READ);



	/*
	 * Get port number for service "remote"
	 */
	
	get_port = getservbyname(REMA_SERVICE, "tcp");
	
	/*
	 * Bind name (port) to local socket
	 */
	
	my_name.sin_family = AF_INET;
	my_name.sin_port = get_port->s_port;
	my_name.sin_addr.s_addr = INADDR_ANY;
	
	setsockopt(in_socket, SOL_SOCKET, SO_REUSEADDR, &on, sizeof (on));
	
	if (bind(in_socket, (struct sockaddr *)&my_name, sizeof my_name) < 0) {
		syslog(LOG_ERR, "Rexec-Manager: Error binding local socket: %m");
		exit(1);
	}  
	
	/*
	 * Get port number for service "ldaemon". We need the loaddaemon for
	 * authentication.
	 */
	
	get_port = getservbyname(DAEMON_SERVICE, "udp");
	
	auth_name.sin_family = AF_INET;
	auth_name.sin_port = get_port->s_port;
	
	/*
	 * Open udp-socket used for authenticcation
	 */
	
	auth_server = socket(AF_INET, SOCK_DGRAM,0);
	
	/*
	 * Init authentication-message
	 */
	
	request.kind = AUTH_REQUEST;
	request.magic_number = MAGIC;
	
	/*
	 * Init result-message
	 */
	
	result.magic_number = MAGIC;
	
	/*
	 * Init notification-message
	 */
	
	notification.magic_number = MAGIC;
	
	/*
	 * Init file descriptor set for sockets with possible oob-data
	 */
	
	FD_ZERO(&exceptfds);
	
	/*
	 * We do not want to exit if we write to an unconnected socket, so
	 * ignore SIGPIPE
	 */
	
	signal (SIGPIPE, SIG_IGN);
	
	/*
	 * Init procs and socks data-structure
	 */
	
	for(count = 0 ; count< MAX_SOCKS; count++)
	  procs[count].proc_id = -1;
	
	for(count = 0; count < MAX_CONNECTS; count++)
	  socks[count].socket = -1;
	
	/*
	 * Mark socket as receiving connections. Parameter backlog not specified
	 */
#if 1
	/* XXX (schoenfr) better define a backlog; C is a wonderful trap... */
	if (listen (in_socket, 10) < 0)
	  {
	    syslog(LOG_ERR, "Rexec-Manager: cannot listen; reason: %m");
	    perror ("Rexec-Manager: cannot listen; reason");
	    exit(-1);
	  }
#else	
	listen (in_socket);
#endif	
	signal(SIGCHLD, handle_SIGCHLD);
	signal(SIGURG, handle_SIGURG);
	
	/*
	 * Now wait for incoming requests of applications and accept them.
	 * According to the SUN Network Programming Guide, the accept-call is
	 * interruptable by a signal, whereas the man page says nothing
	 * about it...
	 */
	
	for(;;){
		
		do {
			
			/*
			 * Blocks indefintely unless signal is delivered or a connection
			 * is accepted...
			 */    
			
			ret = accept (in_socket, (struct sockaddr *)&from, &fromlen);
			
			
		} while ((errno == EINTR) && (ret < 0));
		
		
		if((errno != EINTR) && (ret < 0)) {
			syslog(LOG_ERR, "Rexec-Manager: Could not accept connecton: %m ");
			continue;
		}
		
		/*
		 * Determine free entry in socks data-structure
		 */
		
		for (last = 0; (last < MAX_CONNECTS) && (socks[last].socket >= 0); last++)
		    continue;
		
		/*
		 * Check whether maximum number of connections is exceeded. If so,
		 * shutdown and close socket
		 */
		
		if(socks[last].socket >= 0) {
			
			/*
			 * Send notification of refusal to remote site
			 */
			
			notification.kind = NO_EXECUTION;
			
			/*
			 * Send notification to remote host. We only care for
			 * interruptions due to signals.
			 */
			
			while ((send(ret,&notification, sizeof (notify), 0) < 0)
			       && (errno == EINTR)) ;
			shutdown(ret,2);
			close(ret);
			continue;
		}
		else {
			
			/*
			 * Send notification of acceptance to remote site
			 */
			
			notification.kind = EXECUTION_OK;
			
			/*
			 * Send notification to remote host. We only care for
			 * interruptions due to signals.
			 */
			
			while (send(ret,&notification, sizeof (notify), 0) < 0
			       && errno == EINTR) ;
			
			socks[last].socket = ret;
			socks[last].host = from;
			imported++;
		}
		
		/*
		 * We read the incoming message. To avoid being blocked by a connection
		 * that does not send any request, we use a select with a timeout of
		 * DATA_TIMEOUT microseconds.
		 */
      
		FD_ZERO(&data_fd);
		FD_SET(socks[last].socket,&data_fd);
		
		while((ret = select(MAXWIDTH, &data_fd, (fd_set *)0, (fd_set *)0, &data_timeout)) < 0 && errno == EINTR)
		    continue;
		
		/*
		 * Check whether timeout occured and no data was read. If so, close
		 * the socket and refuse any further requests from the connected socket.
		 * Continue work.
		 */
		
		if (ret <= 0) 
		{
			REFUSE;
			continue;
		}
		
		/*
		 * Otherwise get exec-data from socket. We must be careful at
		 * this point, because it is possible that not the whole
		 * exec_data structure is transmitted at one call of recv
		 */
		
		while ((ret = recv(socks[last].socket, (char *)(&buf),
				   sizeof (buf), 0)) < 0 && errno == EINTR)
		    continue;
		
		if (ret < 0) 
		{
			buf.kind == FALSE;
			syslog(LOG_ERR,
			       "Rexec-Manager: Error reading exec_data : %m ");
		}

		/*
		 * We received a message from remote site. 
		 * Check for correctness of
		 * magic number and message type. If message is incorrect,
		 * close socket, refuse any further requests from the connected
		 * socket and resume main loop.
		 */
		
		if (buf.kind != REQUEST_FOR_EXEC || buf.magic_number != MAGIC) 
		{
			REFUSE;
			syslog(LOG_WARNING, "Rexec-Manager: Job refused...");
			continue;
		} 
		
		if (ret < buf.length)  
		  syslog(LOG_WARNING, 
		     "Rexec-Manager: %d bytes received, %d bytes expected", 
			 ret, buf.length);
		

		/* 
		 * XXX (schoenfr): short messages should be errors, 
		 * not only warnings... 
		 *
		 * XXX should we try to assemble a splitted message ?
		 */


		/*
		 * Incoming request seems to be correct. Now we try to
		 * contact the remote loaddaemon and demand an
		 * authentication of the remote site
		 */
		
		bcopy(&from.sin_addr.s_addr, &auth_name.sin_addr.s_addr,
		      sizeof (from.sin_addr.s_addr));
		
		request.port = from.sin_port;
		request.proc_nr = buf.proc_nr;
		request.sock_nr = buf.sock_nr;
		request.euid = buf.euid;
		request.ruid = buf.ruid;
		request.rgid = buf.rgid;
		request.egid = buf.egid;
		
		for (count=0;count<buf.ngroups;count++)
		  request.groups[count]=buf.groups[count];
		
		/*
		 * Wait for authentication. 
		 */
		
		{
			static stats_timeout timeout_info;

			timeout_info.count=0;
			authenticated = FALSE;
			
			while(timeout_info.count<MAX_TRIES && !authenticated) {
				
				/*
				 * Send request for authentication to remote Loaddaemon
				 */
			
				if (sendto(auth_server, &request, sizeof request,
					   0, (struct sockaddr *) &auth_name,
					   sizeof auth_name) < 0) {
					syslog(LOG_ERR,
			"Rexec-Manager: contacting authentication server: %m");
					continue;
				}
				
				FD_ZERO(&auth_fd);
				FD_SET(auth_server, &auth_fd);
				
				/*
				 * Select for readability on authentication-socket 
				 */
				
				while((ret = select(MAXWIDTH, &auth_fd, (fd_set *)0, (fd_set *)0, &auth_timeout)) < 0 && errno == EINTR);
				
				timeout_info.count++;
				
				if (ret == 0)
				  continue;
				else if (ret < 0) {
					syslog(LOG_ERR,
	       "Rexec-Manager: Error getting authentication from Loaddaemon: %m");
					continue;
				}
				
				if(!FD_ISSET(auth_server, &auth_fd)) {
					continue;
				}
				
				/*
				 * Read Authentication-reply
				 */
				
				while ((ret = recvfrom(auth_server, &reply, 
				          sizeof reply, 0, (struct sockaddr_in *) 
						       &where, &fromlen)) < 0 
				       && errno == EINTR);
				
				if (ret > 0) {
					authenticated = TRUE;
					continue;
				}
				else if (ret < 0) {
					syslog(LOG_ERR,
	      "Rexec-Manager: Error getting authentication from Loaddaemon: %m");
					continue;
				}
			}
			timeout_info.count--;
			if( timeout_info.count > 0 ) {
				/*
				 * Write statistics
				 */
				stats_write(STATS_TYPE_AUTH_TIMEOUT, (stats_info)&timeout_info);
			}
		}
		
		/*
		 * Now we examine incoming reply for correctness
		 */
		
		if(!authenticated || where.sin_port != auth_name.sin_port || reply.kind != AUTH_REPLY ||
		   reply.magic_number != MAGIC || reply.ok != TRUE) {
			ERROR(NO_AUTHENTICATION);
			REFUSE;
			continue;
		}
		
		/*
		 * Make sure the person's not trying to execute as a privileged user 
		 */
		
		if (buf.ruid <= 1024 || buf.euid <= 1024) {
			ERROR(NO_ROOT_EXEC);
			REFUSE;
			continue;
		}
		
		/*
		 * Analyze variable part of exec-data
		 */
		
		analyze = buf.buf;
		
		/*
		 * Get current work directory and perform the one and only
		 * nase_check_path - function on it
		 */
		
		cwd = analyze;
		
		analyze += strlen(cwd) + 1;
		
		if (check_path (cwd, newcwd))
		  cwd = newcwd;
		
		/*
		 * Next data-item. Get the bin-file to execute
		 */
		
		file = analyze;
		
		/*
		 * Next data-item. Extract the arguments to be passed to the file
		 */
		
		analyze += strlen(file) + 1;
		
		/*
		 * Align the resulting pointer to longword boundary and extract argument vectors
		 */
		
		analyze = ALIGN(analyze, char *);
		
		argv = extract (&analyze);
		
		/*
		 * Align the resulting pointer to longword boundary and extract environment vectors
		 */
		
		analyze = ALIGN(analyze, char *);
		
		envp = extract (&analyze);
		
		/*
		 * Set address to connect stdin, stdout, stderr to
		 */
		
		where = from;
		where.sin_port = buf.port;
		
		/*
		 * Set owner of SIGURG to current process and insert socket into
		 * exceptfds to enable selecting for oob-data
		 */
		
		if (fcntl(socks[last].socket, F_SETOWN, getpid()) < 0) {
			ERROR(INTERNAL_ERROR);
			syslog(LOG_ERR, "Rexec-Manager: Error fcntl F_SETOWN: %m");
			exit(1);
		}
		
		FD_SET(socks[last].socket, &exceptfds);
		
		/*
		 * Fork and execute given command 
		 */
		
		(*temp_imported)++;
		
		if( buf.jobkind == LONG_JOB ) {
			/*
			 * Mark the process as a long job and increase the number
			 * of long jobs in shared memory
			 */
			procs[socks[last].socket].long_job = TRUE;
			(*long_jobs)++;
		}
		else {
			procs[socks[last].socket].long_job = FALSE;
		}
		
		gettimeofday(&(procs[socks[last].socket].started), (struct timezone *)0);
		
		main_mask = sigblock(~0);
		
		procs[socks[last].socket].proc_id = fork();
		
		if (procs[socks[last].socket].proc_id == 0) {
			
			/*
			 * Child process: executes given command
			 */
			
			int sock;          /* Socket for stdin, stdout, stderr */
			
			/*
			 * Reinstall standard signal-handlers for SIGURG, SIGCHLD and SIGPIPE
			 */
			
			signal (SIGPIPE, SIG_DFL);
			signal(SIGCHLD, SIG_DFL);
			signal(SIGURG, SIG_DFL);
			
			/*
			 * Unmap the shared memory segment and close attached file
			 */
			
			
			if( munmap(addr_time, (size_t) (sizeof(struct timeval) + sizeof(float) + 2*sizeof(int) + sizeof(float) + sizeof(int)) ) < 0){
				perror ("unmapping shared memory segment");
				exit(-1);
			}
			
			close(shared_mem);
			
			/*
			 * Close all sockets of parent
			 */
			
			close(auth_server);
			close(in_socket);
			
			{ 
				int count;
				for(count=0; count < MAX_CONNECTS; count++) {
					if (socks[count].socket > 2)
					  close(socks[count].socket);
				}
			}
			
			/*
			 * Open socket for stdin, stdout, stderr and connect it to remote site
			 */
			
			sock = socket(AF_INET, SOCK_STREAM, 0);
			
			if (connect(sock, (struct sockaddr *)&where, sizeof (where)) != 0) {
				
				/* 
				 * No correct remote socket specified, so we are in trouble
				 * here. We can't report any error-messages back to our
				 * client (is this our fault ?)...
				 */
				exit(-1);
			}
			
			/*
			 * Redirect the process's stdin/out/err to remote site
			 */
			
			if (sock != 0) {
				dup2 (sock, 0);
			}
			if (sock != 1) {
				dup2 (sock, 1);
			}
			if (sock != 2) {
				dup2 (sock, 2);
			}
			
			if (sock > 2) {
				close(sock);
			}
			
			/*
			 * Set environment pointer
			 */
			
			environ = envp;
			
			/*
			 * Set group-id of user
			 */
			
			if (setregid (buf.rgid, buf.egid) < 0) {
			    /* Error: Could not group-id */
			    perror("Could not set group-id; reason");
			    _exit(-1);
			}
			
#ifdef ROOT_OWNER
			
			/*
			 * Set groups of user
			 */
			
			if (setgroups (buf.ngroups, buf.groups) < 0) {
			    /* Error: Could not set groups */
			    perror ("Could not set groups; reason");
			    _exit (-1);
			}
			
#endif /* ROOT_OWNER */
			
			/*
			 * Set effective user-id of user
			 */
			
			if (setreuid (buf.ruid, buf.euid) < 0) {
				/* Error: Could not set user-id */
				perror("Could not call setreuid; reason");
				_exit (-1);
			}
			
			/*
			 * Change directory to cwd of user
			 */
			
			if (chdir (cwd) < 0) {
			    /* Error: Could not change directory */
			    perror ("Could not change directory; reason");
			    _exit (-1);
			}
			
			/*
			 * Set umask correctly
			 */
			
			umask (buf.umask);
			
			/*
			 * Assign correct process-group to process
			 */
			
			if (setpgrp (0, getpid()) < 0)
			{
			    perror ("Could not call setpgrp; reason");
			    _exit (-1);
			}
			
			/*
			 * Reinstall default sigmask
			 */
			
			sigsetmask(0);
			
			if( buf.jobkind == LONG_JOB && NICE > 0)
			    nice(NICE);
			
			/*
			 * Execute given file
			 */
			execvp (file, argv);

#if 0
			/* oh dear ... */
			free((char *)argv);
			free((char *)envp);
#endif			
			/*
			 * Should not be reached unless error in execution occured 
			 */
			
			/* Error: Could not execute file */			
			perror ("Could not execute file; reason");

			_exit(-1);    /* This is only valid for the child process */
		}
		
		if(procs[socks[last].socket].proc_id < 0) {
			(*temp_imported)--;
			ERROR(NO_FORK);
			REFUSE;
		}
		
		{
			/*
			 * Write statistics
			 */
			static struct hostent *host_ent;
			static stats_job_begin job_begin_info;

			job_begin_info.proc_nr = procs[socks[last].socket].proc_id;
			strcpy(job_begin_info.command, file);
			host_ent = gethostbyaddr((char *)&from.sin_addr.s_addr, sizeof (from.sin_addr.s_addr), AF_INET);
			strcpy(job_begin_info.from, host_ent->h_name); 
			stats_write(STATS_TYPE_JOB_BEGIN, (stats_info)&job_begin_info);
		}
#ifdef PROTOCOL    
		{
			FILE *stats;
			struct hostent *name;
			struct timeval time;
			char host[15];
			
			if(stats =  fopen("/tmp/rexec.log","a")) {
				name = gethostbyaddr((char *)&from.sin_addr.s_addr, sizeof (from.sin_addr.s_addr), AF_INET);
				gethostname(host, sizeof(host));
				gettimeofday(&time, (struct timezone *)0);
				fprintf(stats,"Host: %10s Date: %15s Command received: %10s From host: %10s\n", host, ctime(&time.tv_sec), file, name->h_name);
				fclose(stats);
			}
		}
#endif
		
		free((char *)argv);
		free((char *)envp);
		
		/*
		 * deblock SIGCHLD and SIGURG
		 */
		
		sigsetmask(main_mask);
	}
	
	/*
	 * Never reached
	 */
	exit(0);
}

