/*
 * Copyright (c) 1994
 *
 * Bettina Schnor,  Stefan Stille,  Henrik Carlsson
 *
 * TU Braunschweig, Germany
 * Institute for Operating Systems and Computer Networks
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that this copyright
 * notice appears in all copies.  The University of Braunschweig
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

/*
 * $Id: yalb_system_config.c,v 1.3 1994/03/11 14:35:56 carlsson Exp $
 * $Log: yalb_system_config.c,v $
 * Revision 1.3  1994/03/11  14:35:56  carlsson
 * Added call to `permission()'.
 *
 * Revision 1.2  1994/03/07  10:33:49  carlsson
 * *** empty log message ***
 *
 * Revision 1.1  1994/02/02  10:31:37  carlsson
 * Added errormessages.
 *
 * Revision 1.0  1993/12/21  11:00:19  carlsson
 * Initial revision
 */

#include <stdio.h>
#include "yalb_config_db.h"

#define TRUE 1
#define FALSE 0

#define OK 0
#define ERR_ARGS -1


extern double strtod(char *, char **);
extern long strtol(char *, char **, int);
extern void reconfig(char *);
extern int strcasecmp(char *, char *);
extern int fprintf();

char program[] = "yalb_system_config";

static system_config global_system_config;

static int get_opts(int argc, char *argv[])
{
	int argp = 1;
	char *sp;

	clear_system_config(&global_system_config);
	

	while( argv[argp] ) {
                if( strcasecmp("-timeout", argv[argp]) == 0 && argv[argp + 1] ) {
                        argp++;
			global_system_config.timeout = strtol(argv[argp], &sp, 10);
			if( sp == argv[argp] || global_system_config.timeout < 0 ) {
				return( ERR_ARGS );
			}
                }
                else if( strcasecmp("-polltimeout", argv[argp]) == 0 && argv[argp + 1] ) {
                        argp++;
			global_system_config.poll_timeout = strtol(argv[argp], &sp, 10);
			if( sp == argv[argp] || global_system_config.poll_timeout < 0 ) {
				return( ERR_ARGS );
			}
                }
                else if( strcasecmp("-mintop", argv[argp]) == 0 && argv[argp + 1] ) {
                        argp++;
			global_system_config.minimum_top = (int)strtol(argv[argp], &sp, 10);
			if( sp == argv[argp] || global_system_config.minimum_top < 0 ) {
				return( ERR_ARGS );
			}
                }
		else if( strcasecmp("-idlehost", argv[argp]) == 0 && argv[argp + 1] ) {
			argp++;
			global_system_config.idle_host = (float)strtod(argv[argp], &sp);
			if( sp == argv[argp] || global_system_config.idle_host <= 0.0 ) {
				return( ERR_ARGS );
			}
                }
		else if( strcasecmp("-swapval", argv[argp]) == 0 && argv[argp + 1] ) {
                        argp++;
			global_system_config.swap_validity = (int)strtol(argv[argp], &sp, 10);
			if( sp == argv[argp] || 
			    global_system_config.swap_validity < 0 ||
			    global_system_config.swap_validity > 100 ) {
				return( ERR_ARGS );
			}
                }
		else if( strcasecmp("-improvment", argv[argp]) == 0 && argv[argp + 1] ) {
                        argp++;
			global_system_config.improvment = (int)strtol(argv[argp], &sp, 10);
			if( sp == argv[argp] || 
			    global_system_config.improvment < 0 ||
			    global_system_config.improvment > 100 ) {
				return( ERR_ARGS );
			}
                }
		else if( strcasecmp("-strategy", argv[argp]) == 0 && argv[argp + 1] ) {
			argp++;
			if( strcasecmp(TOPLIST_NAME, argv[argp]) == 0 ) {
				global_system_config.strategy = TOPLIST;
			}
			else if( strcasecmp(TOPHOST_NAME, argv[argp]) == 0 ) {
				global_system_config.strategy = TOPHOST;
			}
			else {
				return( ERR_ARGS );
			}
		}
		else if( strcasecmp("-loadpolling", argv[argp]) == 0 && argv[argp + 1] ) {
			argp++;
			if( strcasecmp(YES_NAME, argv[argp]) == 0 ) {
				global_system_config.loadpolling = TRUE;
			}
			else if( strcasecmp(NO_NAME, argv[argp]) == 0 ) {
				global_system_config.loadpolling = FALSE;
			}
			else {
				return( ERR_ARGS );
			}
		}
		else if( strcasecmp("-statistics", argv[argp]) == 0 && argv[argp + 1] ) {
			argp++;
			if( strcasecmp(YES_NAME, argv[argp]) == 0 ) {
				global_system_config.statistics = TRUE;
			}
			else if( strcasecmp(NO_NAME, argv[argp]) == 0 ) {
				global_system_config.statistics = FALSE;
			}
			else {
				return( ERR_ARGS );
			}
		}
		else {
			return( ERR_ARGS );
		}
		argp++;
	}

        return( OK );
}

static void err(int ret)
{
	switch( ret ) {
	      case ERR_CFDB_OPEN:
		fprintf(stderr, "%s: Can't open configuration file.\n", program);
		break;
	      case ERR_ARGS:
		fprintf(stderr, "usage: %s [-strategy TOPLIST | TOPHOST]\n\t\t\t  [-timeout <timeout>]\n\t\t\t  [-polltimeout <polltimeout>]\n\t\t\t  [-mintop <hosts>]\n\t\t\t  [-idlehost <load>]\n\t\t\t  [-swapval <validity>]\n\t\t\t  [-improvment <level>]\n\t\t\t  [-loadpolling YES | NO]\n\t\t\t  [-statistics YES | NO]\n", program);
	}
	exit(-1);
}

void main(int argc, char *argv[])
{
	int ret;

	permission();

	if( (ret = get_opts(argc, argv)) != OK ) {
                err(ret);
        }
	if( (ret = open_config((int *)NULL)) != OK_CFDB ) {
		err(ret);
	}
	if( (ret = write_system_config(&global_system_config)) != OK_CFDB ) {
		err(ret);
	}
	close_config();

	(void)reconfig_sys();

}
